/*
 *	(c) 1996,1997 - Mathijs Maassen, PE1NTP
 *
 *	v1.0 - 20 June 1996
 *	v1.1 - 23 June 1996
 *	v1.2 - 28 June 1996
 *	v1.3 - 15 July 1996
 *	v1.4 - 10 Dec  1996	(1st Linux version)
 *	v2.0 - 30 April 1997 (Added filter)
 *	v2.02 - 31 Jan 1999 (fixed Linux EOL problem)
 *
 *	(Used TAB width: 4)
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <ctype.h>
#ifdef __MSDOS__
#include <io.h>
#include <dir.h>
#endif
#ifdef __linux__
#include <unistd.h>
#endif
#include "fbb.h"
#include "reqlst.h"

#define REX 50	/* sizeof I/O buffer = REX * sizeof(bullist) */

/*
 *	REQLST.DAT structure
 */
struct lastnum_rec {
	char call[7];
	long lastnum;
};

static char *ver_str = "REQLST v2.02 - (c) 1996,1997,1998,1999 Mathijs Maassen PE1NTP @ PI8DAZ.#TWE.NLD.EU\n";
static char *err_open = "Error opening %s: #%d\n";
static char *lastnumsname = "REQLST/reqlst.dat";
static char *hlpname = "REQLST/reqlst.hlp";
static char *init_srv = "init.srv";
static char *gethlpstr = "Send a message with HELP in the title to get the ReqLst manual.\n";
static char mail_in[80], dirmes[80], user_home[7], to_call[8];

char line[LINELEN];
char user_call[7];
char *server_in;
int line_len = 78, max_lines = 0;

/*
 *	Convert nil terminated string to upper case.
 */
#ifdef NEED_STRUPR
char *strupr(char *buf)
{
	int i;

	for (i = 0; buf[i] != '\0'; i++)
		buf[i] = toupper(buf[i]);
	return (buf);
}
#endif

/*
 *	Cleanup on exit
 */
static void cleanup(void)
{
	if (server_in)
		unlink(server_in);
}

/*
 *	Open file with error checking.
 */
static FILE *open_file(char *path, char *mode)
{
	FILE *fp;

	if ((fp = fopen(path, mode)) == NULL) {
		fprintf(stderr, err_open, path, errno);
		exit(1);
	}
	return (fp);
}

/*
 *	Get pathnames of MAIL.IN and DIRMES.SYS from INIT.SRV
 */
static void init(void)
{
	FILE *fp;
	char *p;
	int lineno = 0;

#ifdef __linux__
	mkdir("REQLST", 0777);
#else
	mkdir("REQLST");
#endif
	fp = open_file(init_srv, "rt");

	while (fgets(line, sizeof(line), fp) != NULL) {
		if (line[0] == '#')
			continue;
		switch (++lineno) {
		case 5:
			strcpy(dirmes, line);
			p = strpbrk(dirmes, " \t\n\r");
			*p = '\0';
			strcat(dirmes, "dirmes.sys");
			continue;
		case 14:
			strcpy(mail_in, line);
			p = strpbrk(mail_in, " \t\n\r");
			*p = '\0';
			break;
		default:
			continue;
		}
		break;
	}

	fclose(fp);
}

/*
 *	Read DIRMES.SYS, buffered I/O, read backwards..
 *	Print list to MAIL.IN file.
 */
static long print_list(FILE *out, char *call, long start)
{
	char date[7], bbs[7], fmt[60], *p;
	long last = 0, cur_pos, bufl = REX * sizeof(bullist), lines = 0;
	bullist *bullbuf;
	int bullp, loop = 1, header = 0, fd;

	if ((bullbuf = malloc(bufl)) == NULL) {	/* I/O buffer */
		fprintf(stderr, "Out of memory.\n");
		exit(1);
	}

	if ((fd = open(dirmes, O_RDONLY|O_BINARY)) < 0) {
		fprintf(stderr, err_open, dirmes, errno);
		exit(1);
	}

	cur_pos = lseek(fd, 0L, SEEK_END);
	bullp = -1;

	while (1) {
		if (bullp < 0) {
			long new_pos;

			if (!loop)
				break;
			if ((new_pos = lseek(fd, -bufl, SEEK_CUR)) <= 0) {	/* Less then REX records left */
				lseek(fd, sizeof(bullist), SEEK_SET);	/* 2nd record */
				new_pos = sizeof(bullist);
				bufl = cur_pos - new_pos;
				bullp = bufl / sizeof(bullist);	/* Number of records left */
				loop = 0;	/* Last round.. */
			} else
				bullp = REX;
			read(fd, bullbuf, bufl);
			cur_pos = lseek(fd, -bufl, SEEK_CUR);
			bullp--;
		}

		if (bullbuf[bullp].numero < start)	/* No new messages */
			break;

		if (bullbuf[bullp].status == 'K' || bullbuf[bullp].status == 'A' || (bullbuf[bullp].type != 'B' && (strcmp(bullbuf[bullp].exped, call) && strcmp(bullbuf[bullp].desti, call)))) {
			bullp--;	/* Next */
			continue;
		}

		if (bullbuf[bullp].type == 'B') {
			if (!test_list_entry(&bullbuf[bullp])) {
				bullp--;
				continue;
			}
		}

#ifdef _MSC_VER
		bullbuf[bullp].date += 2209075200;	/* AAARGH */
#endif
		strftime(date, sizeof(date), "%y%m%d", localtime((time_t *) &bullbuf[bullp].datesd));
		strncpy(bbs, bullbuf[bullp].bbsv, 6);
		if ((p = strchr(bbs, '.')) != NULL)
			*p = '\0';
		else
			bbs[6] = '\0';
		if (!header) {
			header = 1;
			last = bullbuf[bullp].numero + 1;
		}
		sprintf(fmt, "%%6ld %%c %%5ld %%-6s%%s%%-6s %%-6s %%s %%s%%s %%.%ds\n", line_len - 46);
		fprintf(out, fmt, bullbuf[bullp].numero, bullbuf[bullp].type, bullbuf[bullp].taille, bullbuf[bullp].desti, (bbs[0]) ? "@" : " ", bbs, bullbuf[bullp].exped, date, (bullbuf[bullp].free[0] & 1) ? "D" : " ", (bullbuf[bullp].bbsf[0]) ? " " : "L", bullbuf[bullp].titre);
		bullp--;
		if (max_lines && ++lines == max_lines)
			break;
	}

	close(fd);
	free(bullbuf);

	return (last);
}

/*
 *	Complain about syntax errors in filter and exit.
 */
static void check_filter_err(FILE *mailin, char *user, char *home)
{
	if (!flterr[0])
		return;

	fprintf(mailin, "SP %s%s%s\nReqLst: Filter error\n%s\n%s/EX\n",
		user, (*home) ? " @ " : "", home, flterr, gethlpstr);

	fclose(mailin);

	exit(0);
}

/*
 *	Complain about syntax error in subject and exit.
 */
static void usage_error(FILE *mailin, char *user, char *home)
{
	fprintf(mailin, "SP %s%s%s\n"
		"ReqLst error\n"
		"%s/EX\n",
			user, (*home) ? " @ " : "", home, gethlpstr);

	fclose(mailin);

	exit(0);
}

/*
 *	Copy hlpname to mailin and exit.
 */
static void sendhelp(FILE *mailin, char *user, char *home)
{
	FILE *fp;
	int ch;

	fprintf(mailin, "SP %s%s%s\nReqLst help\n", user, (*home) ? " @ " : "", home);

	if ((fp = fopen(hlpname, "r")) == NULL) {
		fprintf(mailin, "Can't open %s!", hlpname);
	} else {
		while ((ch = fgetc(fp)) != EOF)
			fputc(ch, mailin);
	}
	fputs("\n/EX\n", mailin);

	fclose(fp);
	fclose(mailin);

	exit(0);
}

/*
 *	Parse server.in file.
 */
static void get_info(char *fname, char *call, char *home, char *to, long *start)
{
	FILE *fp;
	char temp[21], *p, *p2;

	fp = open_file(fname, "rt");
	fgets(line, sizeof(line), fp);
	sscanf(line, "SP %*s < %6s", call);

	fgets(line, sizeof(line), fp);
	switch (sscanf(line, "%ld %20[^\n\r]", start, temp)) {
	case 1:
		to[0] = '\0';
		home[0] = '\0';
		break;
	case 2:
		if (sscanf(temp, "%7[^@]@%6s", to, home) != 2) {
			to[0] = '\0';
			home[0] = '\0';
			break;
		}
		if ((p = strchr(to, ' ')) != NULL)
			*p = '\0';
		strupr(to);
		strupr(home);
		break;
	default:
		to[0] = '\0';
		home[0] = '\0';
		if (!strncasecmp(line, "SHOW", 4))
			*start = -3;
		else if (!strncasecmp(line, "HELP", 4))
			*start = -2;
		else
			*start = -1;
		fclose(fp);
		return;
	}

	if (!home[0]) {
		long last_pos = ftell(fp);

		while (fgets(line, sizeof(line), fp) != NULL) {
			if (line[0] != 'R' || line[1] != ':') {
				fseek(fp, last_pos, SEEK_SET);
				break;
			}
			last_pos = ftell(fp);
			if ((p = strchr(line, '@')) == NULL)
				break;
			if (*(++p) == ':')
				p++;
			if ((p2 = strpbrk(p, ".\t\n\r ")) == NULL)
				break;
			*p2 = '\0';
			strcpy(home, p);
		}
	}

	if (init_filter(fp) == 0)
		restore_saved_filter(call);

	fclose(fp);
}

/*
 *	Update last listed number in REQLST.DAT
 */
static void update_lastnum(char *call, long last)
{
	FILE *fp;
	struct lastnum_rec rec;

	fclose(open_file(lastnumsname, "a"));
	fp = open_file(lastnumsname, "r+b");
	while (fread(&rec, sizeof(rec), 1, fp) > 0) {
		if (!strcmp(call, rec.call)) {
			fseek(fp, (int) -sizeof(rec), SEEK_CUR);
			rec.lastnum = last;
			fwrite(&rec, sizeof(rec), 1, fp);
			fclose(fp);
			return;
		}
	}

	strcpy(rec.call, call);
	rec.lastnum = last;
	fwrite(&rec, sizeof(rec), 1, fp);

	fclose(fp);
}

/*
 *	Get last listed number from REQLST.DAT
 */
static long get_lastnum(char *call)
{
	FILE *fp;
	struct lastnum_rec rec;

	if ((fp = fopen(lastnumsname, "rb")) == NULL)
		return (0);
	while (fread(&rec, sizeof(rec), 1, fp) > 0) {
		if (!strcmp(call, rec.call)) {
			fclose(fp);
			return (rec.lastnum);
		}
	}

	fclose(fp);

	return (0);
}

int main(int argc, char **argv)
{
	long start_num, last_num;
	FILE *outfp;

	if (argc != 2) {
		fprintf(stderr, ver_str);
		return (1);
	}

	init();

	server_in = argv[1];
	atexit(cleanup);

#ifdef DEBUG
	outfp = stdout;
#else
	outfp = open_file(mail_in, "at");
#endif /* DEBUG */

	get_info(argv[1], user_call, user_home, to_call, &start_num);
	switch (start_num) {
	case 0:
		start_num = get_lastnum(user_call);
		break;
	case -1:
		usage_error(outfp, user_call, user_home);
		return (0);
	case -2:
		sendhelp(outfp, user_call, user_home);
		return (0);
	case -3:
		show_filt = 1;
		last_num = -1;
		break;
	}

	check_filter_err(outfp, user_call, user_home);

#ifndef __linux__				/* default buffer big enough for Linux */
	setvbuf(outfp, NULL, _IOFBF, 8192);
#endif /* __linux__ */

	if (start_num >= 0) {
		fprintf(outfp, "#\nSP %s%s%s < REQLST\n"
			"ReqLst reply\n", (to_call[0]) ? to_call : user_call, (user_home[0]) ? " @ " : "", user_home);
		last_num = print_list(outfp, user_call, start_num);
		fprintf(outfp, "/EX\n");
	} else
		last_num = -1;

	if (save_filt || del_filt) {
		fprintf(outfp, "#\nSP %s%s%s\n"
			"ReqLst: command confirmation\n"
			"Your filter has been %sed.\n"
			"/EX\n", (to_call[0]) ? to_call : user_call, (user_home[0]) ? " @ " : "", user_home, (del_filt) ? "delet" : "sav");
	}

	if (show_filt) {
		FILE *fp;
		char fname[80] = "REQLST/";
		int ch;

		strcat(fname, user_call);
		strcat(fname, ".flt");
		fp = fopen(fname, "rt");

		fprintf(outfp, "#\nSP %s%s%s\n"
			"ReqLst: your current filter\n", (to_call[0]) ? to_call : user_call, (user_home[0]) ? " @ " : "", user_home);
		if (fp == NULL)
			fputs("No filter found.\n", outfp);
		else {
			while ((ch = fgetc(fp)) != EOF)
				fputc(ch, outfp);
			fclose(fp);
		}
		fprintf(outfp, "/EX\n");
	}

#ifndef DEBUG
	fclose(outfp);
#endif /* DEBUG */

	if (last_num >= 0)
		update_lastnum(user_call, last_num);

	return (0);
}
